// SPDX-License-Identifier: GPL-2.0-or-later
// Copyright The Music Player Daemon Project

#ifndef MPD_CDIO_PARANOIA_INPUT_PLUGIN_HXX
#define MPD_CDIO_PARANOIA_INPUT_PLUGIN_HXX

/**
 * An input plugin based on libcdio_paranoia library.
 */
extern const struct InputPlugin input_plugin_cdio_paranoia;

#endif
