// SPDX-License-Identifier: GPL-2.0-or-later
// Copyright The Music Player Daemon Project

#ifndef MPD_ARCHIVE_BZ2_HXX
#define MPD_ARCHIVE_BZ2_HXX

struct ArchivePlugin;

extern const ArchivePlugin bz2_archive_plugin;

#endif
