// SPDX-License-Identifier: GPL-2.0-or-later
// Copyright The Music Player Daemon Project

#ifndef MPD_SERIAL_HXX
#define MPD_SERIAL_HXX

/**
 * Generate the next pseudo-random serial.
 */
int
GenerateSerial() noexcept;

#endif
