// SPDX-License-Identifier: GPL-2.0-or-later
// Copyright The Music Player Daemon Project

#ifndef MPD_DECODER_GME_HXX
#define MPD_DECODER_GME_HXX

extern const struct DecoderPlugin gme_decoder_plugin;

#endif
