// SPDX-License-Identifier: GPL-2.0-or-later
// Copyright The Music Player Daemon Project

#ifndef MPD_UPNP_TAGS_HXX
#define MPD_UPNP_TAGS_HXX

/**
 * Map UPnP property names to MPD tags.
 */
extern const struct tag_table upnp_tags[];

#endif
