// SPDX-License-Identifier: GPL-2.0-or-later
// Copyright The Music Player Daemon Project

#ifndef MPD_ALSA_INPUT_PLUGIN_HXX
#define MPD_ALSA_INPUT_PLUGIN_HXX

#include "../InputPlugin.hxx"

extern const struct InputPlugin input_plugin_alsa;


#endif
