// SPDX-License-Identifier: GPL-2.0-or-later
// Copyright The Music Player Daemon Project

#ifndef MPD_NULL_FILTER_PLUGIN_HXX
#define MPD_NULL_FILTER_PLUGIN_HXX

struct FilterPlugin;

extern const FilterPlugin null_filter_plugin;

#endif
