// SPDX-License-Identifier: GPL-2.0-or-later
// Copyright The Music Player Daemon Project

#ifndef MPD_DECODER_OPUS_H
#define MPD_DECODER_OPUS_H

extern const struct DecoderPlugin opus_decoder_plugin;

#endif
